package eu.jonahbauer.raytracing.math;

import eu.jonahbauer.raytracing.render.spectrum.SampledWavelengths;
import eu.jonahbauer.raytracing.scene.HitResult;
import org.jetbrains.annotations.NotNull;

import java.util.Objects;

public final class Ray {
    private final @NotNull Vec3 origin;
    private final @NotNull Vec3 direction;
    private final @NotNull SampledWavelengths lambda;

    private final @NotNull Vec3 inv;
    private final @NotNull Vec3 negInvOrigin;

    public Ray(@NotNull Vec3 origin, @NotNull Vec3 direction) {
        this(origin, direction, SampledWavelengths.EMPTY);
    }

    public Ray(@NotNull Vec3 origin, @NotNull Vec3 direction, @NotNull SampledWavelengths lambda) {
        this.origin = Objects.requireNonNull(origin, "origin");
        this.direction = Objects.requireNonNull(direction, "direction");
        this.lambda = Objects.requireNonNull(lambda, "lambda");

        this.inv = direction.inv();
        this.negInvOrigin = inv.neg().times(origin);
    }

    private Ray(@NotNull Vec3 origin, @NotNull Vec3 direction, @NotNull SampledWavelengths lambda, @NotNull Vec3 inv, @NotNull Vec3 negInvOrigin) {
        this.origin = origin;
        this.direction = direction;
        this.lambda = lambda;
        this.inv = inv;
        this.negInvOrigin = negInvOrigin;
    }

    public @NotNull Vec3 at(double t) {
        return Vec3.fma(t, direction, origin);
    }

    public @NotNull Ray with(@NotNull HitResult hit, @NotNull Vec3 direction) {
        return new Ray(hit.position(), direction, lambda);
    }

    public @NotNull Ray with(@NotNull Vec3 origin, @NotNull Vec3 direction) {
        return new Ray(origin, direction, lambda);
    }

    public @NotNull Ray with(@NotNull SampledWavelengths lambda) {
        return new Ray(origin, direction, lambda, inv, negInvOrigin);
    }

    public @NotNull Vec3 origin() {
        return origin;
    }

    public @NotNull Vec3 direction() {
        return direction;
    }

    public @NotNull SampledWavelengths lambda() {
        return lambda;
    }

    public @NotNull Vec3 getInvDirection() {
        return inv;
    }

    public @NotNull Vec3 getNegInvOrigin() {
        return negInvOrigin;
    }

    @Override
    public boolean equals(Object obj) {
        if (obj == this) return true;
        if (obj == null || obj.getClass() != this.getClass()) return false;
        var that = (Ray) obj;
        return Objects.equals(this.origin, that.origin) &&
                Objects.equals(this.direction, that.direction) &&
                Objects.equals(this.lambda, that.lambda);
    }

    @Override
    public int hashCode() {
        return Objects.hash(origin, direction, lambda);
    }

    @Override
    public @NotNull String toString() {
        return "Ray[" +
                "origin=" + origin + ", " +
                "direction=" + direction + ", " +
                "lambda=" + lambda + ']';
    }

}
